import { deepEqual } from "./deep.js";
import { LoggerConsoleHandler } from "./log-console.js";
import { useNamespaceFilter } from "./log-filter.js";
export var LogLevel;
(function (LogLevel) {
    LogLevel[LogLevel["debug"] = 0] = "debug";
    LogLevel[LogLevel["info"] = 1] = "info";
    LogLevel[LogLevel["warn"] = 2] = "warn";
    LogLevel[LogLevel["error"] = 3] = "error";
    LogLevel[LogLevel["fatal"] = 4] = "fatal";
})(LogLevel || (LogLevel = {}));
export function LoggerContext(prefix = "") {
    let logHandlers = [LoggerConsoleHandler()];
    let logAssertLevel = LogLevel.warn;
    let logCheckNamespace = (name) => true;
    let logLock = false;
    let logFactory = LoggerBaseFactory;
    function LoggerBaseFactory(name = "") {
        log.extend = function (prefix) {
            return logFactory(name ? `${name}:${prefix}` : prefix);
        };
        const emit = (msg) => {
            if (log.active === true) {
                if (msg.level >= Logger.level && msg.level >= log.level) {
                    if (logCheckNamespace(name)) {
                        for (let handler of logHandlers) {
                            if (handler)
                                handler(msg);
                        }
                    }
                }
            }
        };
        function log(...messages) {
            emit({
                name,
                messages,
                level: LogLevel.debug,
            });
        }
        log.active = true;
        log.level = LogLevel.debug;
        log.debug = function (...messages) {
            emit({
                name,
                messages,
                level: LogLevel.debug,
            });
        };
        log.info = function (...messages) {
            emit({
                name,
                messages,
                level: LogLevel.info,
            });
        };
        log.warn = function (...messages) {
            emit({
                name,
                messages,
                level: LogLevel.warn,
            });
        };
        log.error = function (...messages) {
            emit({
                name,
                messages,
                level: LogLevel.error,
            });
        };
        // fatal(...args: any[]) {
        //   console.error(...args)
        // },
        log.assert = function (cond, ...messages) {
            if (!cond) {
                if (typeof console !== undefined) {
                    if (console.assert) {
                        // https://developer.mozilla.org/de/docs/Web/API/Console/assert
                        console.assert(cond, ...messages);
                    }
                    else {
                        console.error(`Assert did fail with: ${cond}`, ...messages);
                    }
                }
                emit({
                    name,
                    messages: messages || [`Assert did fail with: ${cond}`],
                    level: logAssertLevel,
                });
                // try {
                //   if (typeof expect !== undefined) {
                //     expect(cond).toBeTruthy()
                //   }
                // } catch (err) {
                //   methods.warn(...args)
                // }
            }
        };
        log.assertEqual = function (value, expected, ...args) {
            let equal = deepEqual(value, expected);
            if (!equal) {
                log.assert(equal, `Assert did fail. Expected ${expected} got ${value}`, expected, value, ...args);
                // } else {
                //   methods.debug(`Passed equal`)
            }
        };
        log.assertNotEqual = function (value, expected, ...args) {
            let equal = deepEqual(value, expected);
            if (equal) {
                log.assert(equal, `Assert did fail. Expected ${expected} not to be equal with ${value}`, expected, value, ...args);
                // } else {
                //   methods.debug(`Passed not equal check`)
            }
        };
        return log;
    }
    function Logger(name = "") {
        return logFactory(name);
    }
    Logger.registerHandler = function (handler) {
        logHandlers.push(handler);
    };
    Logger.setFilter = function (namespaces) {
        logCheckNamespace = useNamespaceFilter(namespaces);
    };
    Logger.setLock = (lock = true) => (logLock = lock);
    Logger.setHandlers = function (handlers = []) {
        if (logLock)
            return;
        logHandlers = [...handlers].filter((h) => typeof h === "function");
    };
    Logger.level = LogLevel.debug;
    /** @deprecated */
    Logger.setLogLevel = function (level = 0) {
        if (logLock)
            return;
        Logger.level = level;
    };
    Logger.setFactory = function (factory) {
        if (logLock)
            return;
        logFactory = factory;
    };
    return Logger;
}
export const Logger = LoggerContext();
let globalLogger = Logger;
try {
    if (typeof window != null) {
        if (window._zeedGlobalLogger == null) {
            window._zeedGlobalLogger = Logger;
        }
        else {
            globalLogger = window._zeedGlobalLogger;
        }
    }
}
catch (e) { }
export const GlobalLogger = globalLogger;
//# sourceMappingURL=log.js.map